#include "Viewer.h"

#include <qcompleter.h>
#include <qdir.h>
#include <qfiledialog.h>
#include <qdebug.h>
#include <qtextdocument.h>
#include <qtexttable.h>

#pragma hdrstop

namespace dbase {

  Viewer::Viewer( QWidget* parent )
    : QWidget( parent )
    , _cacheSize( 0 )
  {
    ui.setupUi( this );

    ui.databaseSearch->setCompleter( new QCompleter( ui.databases->model() ) );

    connect( ui.databases, SIGNAL( itemClicked( QListWidgetItem* ) ), this, SLOT( itemClicked( QListWidgetItem* ) ) );
    connect( ui.databaseSearch, SIGNAL( textChanged( const QString& ) ), this, SLOT( searchTextChanged( const QString& ) ) );
    connect( ui.deleteCache, SIGNAL( clicked() ), this, SLOT( deleteCache() ) );
    connect( ui.writeSchema, SIGNAL( clicked() ), this, SLOT( writeSchema() ) );
    connect( ui.removeEmptyFiles, SIGNAL( stateChanged ( int ) ), this, SLOT( removeEmptyFiles( int ) ) );
  }

  void Viewer::setDirectory( const QString& searchDir ) {
    ui.dir->setText( searchDir );
    ui.databases->clear();

    QDir dir( searchDir );
    dir.setNameFilters( QStringList( "*.dbf" ) );
    QStringList files = dir.entryList();

    QProgressDialog* dlg = new QProgressDialog( tr("Initialisiere dbase Viewer"), QString(), 0, files.count(), this );
    dlg->show();

    QListWidgetItem* itm = NULL;
    bool ok = false;
    for(int i=0; i < files.count(); ++i) {
      itm = new QListWidgetItem( files.at( i ) );
      QString filePath = dir.absoluteFilePath( files.at( i ) );
      if( dbase::DBASE::isEmpty( filePath, ok ) ) {
        itm->setBackgroundColor( ok ? Qt::gray : Qt::darkRed );
      }

      itm->setData( Qt::UserRole, filePath );
      ui.databases->addItem( itm );

      dlg->setValue( i );
    }

    if( files.empty() ) {
      itm = new QListWidgetItem( tr("Keine Daten gefunden." ) );
      itm->setFlags( itm->flags() & ~Qt::ItemIsSelectable );
      ui.databases->addItem( itm );
    }

    dlg->close();
  }

  void Viewer::searchTextChanged( const QString& text ) {
    QList< QListWidgetItem* > items = ui.databases->findItems( text, Qt::MatchContains );
    
    if( items.isEmpty() ) {
      return;
    }

    ui.databases->setCurrentItem( items.at( 0 ) );
    //itemClicked( items.at( 0 ) );
  }

  void Viewer::itemClicked( QListWidgetItem* item ) {
    QString file = item->data( Qt::UserRole ).toString();

    QMap< QString, TableModel* >::iterator itr = _models.find( file );

    TableModel* model = NULL;
    if( itr == _models.end() ) {
      QProgressDialog* dlg = new QProgressDialog();
      dlg->setCancelButton( NULL );
      dlg->show();
      DBASE* db = DBASE::readFrom( file, dlg );
      dlg->close();

      if( db != NULL ) {
        model = new TableModel( db, this );
        _models.insert( file, model );
        
        highlight( item, true );

        _cacheSize += db->getInfo()._size;
        updateCacheProgressBar();

        if( ui.progressBarCache->value() >= ui.progressBarCache->maximum() ) {
          _cacheSize += db->getInfo()._size;
          deleteCache( false );
          updateCacheProgressBar();
        }
      }
    } else {
      model = itr.value();
    }

    if( model == NULL || !model->isValid() ) {
      ui.lVersion->setText( "???" );
      ui.lUpdate->setText( "???" );
      ui.lSize->setText( "???" );

      ui.cbDBT->setChecked( false );
      ui.cbMemo->setChecked( false );
      ui.cbSQL->setChecked( false );

      highlight( item, false );

      ui.tableView->setModel( NULL );
      return;
    }

    ui.tableView->setModel( model );
    const DBASEInfo& info = model->getDBASE()->getInfo();

    ui.lVersion->setText( QString::number( info._versionNumber ) + " (" + QString::number( info._raw ) + ")" );
    ui.lUpdate->setText( info._lastUpdate.toString( Qt::SystemLocaleShortDate ) );

    ui.cbDBT->setChecked( info._dbt );
    ui.cbMemo->setChecked( info._memoFilePresence );
    ui.cbSQL->setChecked( info._sqlTablePresence );

    QString ls = "Bytes";
    qint64 s = info._size;
    if( info._size > 1024 ) {
      ls = "KB";
      s /= 1024;
      if( s > 1024 ) {
        ls = "MB";
        s /= 1024;
      }
    }

    ui.lSize->setText( "~" + QString::number( s ) + " " + ls );
  }

  void Viewer::deleteCache(bool reloadCurrent) {
    for(QMap< QString, TableModel* >::iterator itr = _models.begin(); itr != _models.end(); ++itr) {
      TableModel* cur = itr.value();
      cur->setParent( NULL );
      delete cur;
    }

    _models.clear();

    for(int i=0; i < ui.databases->count(); ++i) {
      highlight( ui.databases->item( i ), false );
    }

    if( reloadCurrent ) {
      itemClicked( ui.databases->currentItem() );
    }
    _cacheSize = 0;
    updateCacheProgressBar();
  }

  void Viewer::removeEmptyFiles(int s) {
    if( s != Qt::Checked ) {
      for(int i=0; i < ui.databases->count(); ++i) {
        ui.databases->setRowHidden( i, false );       
      }      
      return;
    }

    for(int i=ui.databases->count()-1; i >= 0; --i) {
      QListWidgetItem* item = ui.databases->item( i );
      if( item->backgroundColor() == Qt::gray
        || item->backgroundColor() == Qt::darkRed ) {
        ui.databases->setRowHidden( i, true );       
      }
    }
  }

  void Viewer::writeSchema() {
    QString dir = QFileDialog::getExistingDirectory( this, "Speichern unter ..." );
    
    if( dir.isEmpty() ) {
      qCritical() << "Dateiname ist leer, daher kann keine Struktur rausgeschrieben werden.";
      return;
    }    
    
    QProgressDialog* dlg = new QProgressDialog( tr("Schreibe Datenstrukturen raus"), QString(), 0, ui.databases->count(), this );
    dlg->show();

    QTextCharFormat tableName;
    QFont tableNameFont;
    tableNameFont.setBold( true );
    tableNameFont.setPointSize( tableNameFont.pointSize() * 2 );
    tableName.setFont( tableNameFont );

    QTextTableCellFormat tableColTitle;
    QFont tableColTitleFont;
    tableNameFont.setBold( true );
    tableColTitle.setFont( tableColTitleFont );

    QListWidgetItem* itm = NULL;
    QString filepath;
    for(int i=0; i < ui.databases->count(); ++i) {
      itm = ui.databases->item( i );
      filepath = itm->data( Qt::UserRole ).toString();
      dlg->setValue( i );
      if( itm->isHidden() ) {
        qDebug() << "Folgende Datei wurde bersprungen, da ausgeblendet: " << filepath;
        continue;
      }
      
      std::auto_ptr< dbase::DBASE > dbase( dbase::DBASE::readFrom( filepath ) );
      if( dbase.get() == NULL ) {
        qWarning() << "Konnte Datei '" << filepath << "' nicht lesen";
        continue;
      }

      const DBASE::Records& records = dbase->getRecords();
      if( records.isEmpty() ) {
        qDebug() << "Folgende Datei wurde bersprungen, da keine Daten vorhanden: " << filepath;
        continue;
      }

      const DBASE::Fields& fields = dbase->getFields();
      QVector< int > fieldIdsWithData;
      fieldIdsWithData.reserve( fields.size() );

      for(int r=0; r < records.count(); ++r) {
        if( fieldIdsWithData.count() == fields.size() ) {
          break;
        }

        for(DBASE::Fields::size_type f=0; f < fields.size(); ++f) {
          if( fieldIdsWithData.contains( f ) ) {
            continue;
          }

          if( !records.at( r ).get( f ).isNull() ) {
            fieldIdsWithData.push_back( f );
          }
        }
      }

      //
      QFileInfo fileInfo( filepath );      
      QFile outputFile( dir + "/" + fileInfo.baseName() + ".html" );
      if( !outputFile.open( QIODevice::WriteOnly ) ) {
        qCritical() << "Kann Datei nicht zum schreiben ffnen. " << outputFile.fileName();
        return;
      }
      QTextStream output( &outputFile );

      QTextDocument* document = new QTextDocument();
      QTextCursor* cursor = new QTextCursor( document );

      cursor->insertText( itm->text() + " Anzahl Datenstze: " + QString::number( records.count() ), tableName );
      QTextTable* table = cursor->insertTable( fieldIdsWithData.count() + 1, records.count() > 1 ? 3 : 2 );
      table->cellAt( 0, 0 ).setFormat( tableColTitle );
      table->cellAt( 0, 1 ).setFormat( tableColTitle );

      if( records.count() > 1 ) { 
        table->cellAt( 0, 2 ).setFormat( tableColTitle );
      }

      cursor->insertText( "Spaltenname" );
      cursor->movePosition( QTextCursor::NextCell );

      cursor->insertText( "Bsp.Inhalt-0" );
      cursor->movePosition( QTextCursor::NextCell );

      if( records.count() > 1 ) { 
        cursor->insertText( "Bsp.Inhalt-1" );
        cursor->movePosition( QTextCursor::NextCell );
      }

      // schreibe alle felder raus, die in min. einem Datensatz genutzt werden
      for(int j=0; j < fieldIdsWithData.count(); ++j) {
        int fieldId = fieldIdsWithData.at( j );

        cursor->insertText( fields[ fieldId ].getName() );
        cursor->movePosition( QTextCursor::NextCell );

        cursor->insertText( records.at( 0 ).get( fieldId ).toString() );
        cursor->movePosition( QTextCursor::NextCell );

        if( records.count() > 1 ) { 
          cursor->insertText( records.at( records.count() - 1 ).get( fieldId ).toString() );
          cursor->movePosition( QTextCursor::NextCell );
        }        
      }

      cursor->movePosition( QTextCursor::End );
      output << document->toHtml();    
      outputFile.close();
    }    
    
    dlg->close();
  }
}