#include "dbase.h"

#include "qfile.h"
#include <string>

#include <qdebug.h>
#include <qtextcodec.h>
#include <qfileinfo.h>

#pragma hdrstop

namespace dbase {

  bool DBASE::isEmpty( const QString& filename, bool& ok ) {
    ok = false;
    QFile file( filename );
    if( !file.open( QIODevice::ReadOnly ) ) {
      qWarning() << "Kann die Datei '" << filename << "' nicht ffnen.";
      return true;
    }

    // read header
    std::auto_ptr< dbase::raw::Header > header = readHeader( file );
    

    if( header.get() != NULL ) {
      ok = true;
      if( header->_numberOfRecords == 1 && file.seek( header->_headerSize ) ) {
        char a;
        file.read( &a, 1 );
        if( a == 0x1A ) {
          file.close();
          return true;
        } else if( a != raw::Record::VALID && a != raw::Record::DELETED ) {
          qCritical() << filename << ": Ungltiger Wert wurde gefunden." ;
        }
      }
    }

    file.close();
    return header.get() == NULL || header->_numberOfRecords == 0;
  }
  
  DBASE* DBASE::readFrom( const QString& filename, QProgressDialog* dlg /*= NULL*/  ) {
    QFile file( filename );
    if( !file.open( QIODevice::ReadOnly ) ) {
      qWarning() << "Kann die Datei '" << filename << "' nicht ffnen.";
      return NULL;
    }

    if( dlg != NULL ) {
      dlg->setLabelText( "Lese allgemeine Header-Informationen ein." );
      dlg->setMinimum( 0 );
      dlg->setMaximum( file.size() );
    }

    qDebug() << "----";
    qDebug() << "Datei: " << filename;
    qDebug() << "Dateigre: " << file.size() << " Bytes";
    qDebug() << "Header: " << sizeof( dbase::raw::Header ) << " Bytes";
    qDebug() << "FieldDescriptor: " << sizeof( dbase::raw::FieldDescriptor ) << " Bytes";

    // read header
    std::auto_ptr< dbase::raw::Header > header = readHeader( file );

    if( header.get() == NULL ) {
      qCritical() << "Kann den Header der Datei '" << filename << "' nicht lesen.";
      return NULL;
    }

    if( file.pos() != 32 ) {
      qCritical() << "Dateiposition nach lesen des Headers ist fehlerhaft!";
      return NULL;
    }

    //Length of header + ( Number of records * Length of each record
    const qint64 SUM_RECORD_SIZE = header->_numberOfRecords * header->_recordSize ;

    if( dlg != NULL ) {
      dlg->setLabelText( "Lese Spalteninformationen ein." );
      dlg->setValue( file.pos() );
    }

    // read field descriptors
    std::vector< dbase::raw::FieldDescriptor > fields;
    if( !readFieldDescriptor( file, header.get(), fields ) ) {
      qCritical() << "Kann die Feldinformationen der Datei '" << filename << "' nicht lesen.";
      return new DBASE( filename, header.get() );
    }

    // debug
    int recordSizeDebug = 1;
    for( std::vector< dbase::raw::FieldDescriptor >::size_type i=0; i < fields.size(); ++i ) {
      dbase::raw::FieldDescriptor& cur = fields[ i ];
      recordSizeDebug += cur._length;
    }

    if( recordSizeDebug != header->_recordSize ) {
      qCritical() << "Information ber die Datensatzgre ist ungltig. Datensatzgre: '" << header->_recordSize << "' errechnete Gre: '" << recordSizeDebug << "'";
      return NULL;
    } else {
      qDebug() << "> Information zur Datensatzgre ist gltig.";
    }

    qDebug() << "Datensatzgre: " << recordSizeDebug << " Bytes";

    // read terminator 0x0D
    char terminator;
    qint64 tmp = file.read( &terminator, sizeof( char ) );

    if( terminator != 0x0D ) {
      qCritical() << "Gelesen: " << file.pos() << " Bytes Mglich: " << file.size() << "(" << QString::number( file.pos() / (double)file.size() * 100.0, 'f', 1 ) << " %)";
      if( tmp != 1 ) {
        qCritical() << "Das Abschlussbyte nach der Defintion der Felder konnte nicht ausgelesen werden ('" << filename << "')";
      }

      qCritical() << "Das Abschlussbyte nach der Defintion der Felder ist fehlerhaft. ('" << filename << "')";
      return NULL;
    } else {
      qDebug() << "> Abschlussbyte nach Felddefinitionen ist korrekt.";
    }


    if( header->_version & 0x30 ) {
      // read database container
      char tmpChar[ 263 ];
      tmp = file.read( &tmpChar[0], 263 );

      if( tmp != 263 ) {
        qCritical() << "Der Datenbankcontainer konnte nicht vollstndig ausgelesen werden. Gelesen: " << tmp << " Bytes Erfordert: 263 Bytes";
        return NULL;
      } else {
        qDebug() << "> Datenbankcontainer konnte gelesen werden.";
      } 
    }

    // read records
    // check logical size
    qint64 availableBytesToRead = file.size() - file.pos();
    if( availableBytesToRead < SUM_RECORD_SIZE+1 ) {
      qCritical() << "Die logische Datensatzblock ist grer (" << SUM_RECORD_SIZE << "+1 Bytes) als die noch zur Verffung stehenden Bytes ("<< availableBytesToRead << " Bytes).";
      return NULL;
    } 

    if( dlg != NULL ) {
      dlg->setLabelText( "Lese Datenstze ein." );
      dlg->setValue( file.pos() );
    }

    dbase::raw::Record* records = readRecords( file, header.get() );

    if( header->_numberOfRecords != 0 && records == NULL ) {
      qCritical() << "Die Datenbankeintrge konnte nicht gelesen werden. ('" << filename << "')";
      return NULL;
    }

    char eof;
    file.read( &eof, sizeof( char ) );

    if( eof != 0x1A ) {
      qCritical() << "Daten schliet nicht mit dem EOF-Marker ab.";
    }

    if( dlg != NULL ) {
      dlg->setLabelText( "Fertig mit einlesen, Objekt wird erzeugt." );
      dlg->setValue( file.pos() );
    }

    // close file
    file.close();

    // generate DBASE object
    return new DBASE( filename, header.get(), fields, records );
  }

  std::auto_ptr< dbase::raw::Header > DBASE::readHeader( QFile& file ) {
    std::auto_ptr< dbase::raw::Header > header( new dbase::raw::Header() );
    qint64 rs = file.read( (char*)header.get(), sizeof( dbase::raw::Header ) );

    if( rs != sizeof( dbase::raw::Header ) ) {
      qCritical() << "Es konnte nicht der komplette Header gelesen werden. Gelesen: " << rs << " Bytes Headergre: " << sizeof( dbase::raw::Header ) << " Bytes";
    }

    return header;
  }

  bool DBASE::readFieldDescriptor( QFile& file, const dbase::raw::Header* header, std::vector< dbase::raw::FieldDescriptor >& res ) {
    if( file.pos() != 32 ) {
      qCritical() << "Es knnen keine Feldinformationen eingelesen werden, da die aktuelle Dateiposition stimmt nicht.";
      return false;
    }

    int sumOfFields = ( header->_headerSize - sizeof( dbase::raw::Header ) - sizeof( char ) ) / sizeof( dbase::raw::FieldDescriptor );
    res.resize( sumOfFields );

    qDebug() << "- Einlesen der Feldinformationen";

    for(int i=0; i < (int)sumOfFields; ++i) {
      qint64 rs = file.read( (char*)&res[i], sizeof( dbase::raw::FieldDescriptor) );

      if( rs != sizeof( dbase::raw::FieldDescriptor) ) {
        qCritical() << i << ": Es konnte die Feldinformation nicht eingelesen werden.";
      }

      if( file.pos() % 32 != 0 ) {
        qCritical() << i << ": Dateiposition ist nicht durch 32 teilbar.";
      }
    }

    return true;
  }

  dbase::raw::Record* DBASE::readRecords( QFile& file, const dbase::raw::Header* header ) {
    qDebug() << "- Einlesen der Datenstze";

    if( !file.seek( header->_headerSize ) ) {
      qCritical() << "Konnte die Leseposition nicht nach dem Header setzen.";
    }

    if( header->_numberOfRecords == 0 ) {
      qDebug() << "Es stehen keine Datenstze zur Verfgung.";
      return NULL;
    }

    dbase::raw::Record* res = new dbase::raw::Record[ header->_numberOfRecords ];

    const int DATA_SIZE = header->_recordSize - sizeof( dbase::raw::Record::State );

    qDebug() << "Eigentliche Datensatzgre: " << DATA_SIZE << " Bytes";

    for(unsigned int i=0; i < header->_numberOfRecords; ++i) {
      dbase::raw::Record* cur = &res[ i ];
      qint64 curReadSize = file.read( (char*)&cur->_state, sizeof( dbase::raw::Record::State ) );

      if( curReadSize != sizeof( dbase::raw::Record::State ) ) {
        qCritical() << "Datensatz[ " << i << "] Es konnte nicht den Datensatzzustand gelesen werden. Gelesen: " << curReadSize << " Bytes Erfordert: " << sizeof( dbase::raw::Record::State ) << " Bytes";
      }

      cur->_data = new char[ DATA_SIZE ];
      curReadSize = file.read( cur->_data, DATA_SIZE );
      if( curReadSize != DATA_SIZE ) {
        qCritical() << "Datensatz: Konnte nicht soviel gelesen werden, wie erforderlich ist. Gelesen: " << curReadSize << " Bytes Angefordert: " << DATA_SIZE << " Bytes";
      }
    }

    return res;
  }

  void DBASE::initInfo( const dbase::raw::Header* header ) {
    _info._raw = header->_version;

    _info._versionNumber = _info._raw & 7;
    _info._memoFilePresence = _info._raw & 8;
    _info._sqlTablePresence = _info._raw & 16 || _info._raw & 32 || _info._raw & 64;
    _info._dbt = _info._raw & 128;

    int y = 2000 + header->_lastUpdateYear;
    int m = header->_lastUpdateMonth;
    int d = header->_lastUpdateDay;

    _info._lastUpdate = QDate( y, m, d );
    _info._size = header->_headerSize + header->_numberOfRecords * header->_recordSize;
  }

  QVariant::Type DBASE::toVariantType(const dbase::raw::FieldDescriptor& desc ) const {
    /* 

      DATE_TIME = 'T',
      INTEGER = 'I',
      VARI_FIELD = 'V',
      VARIANT = 'X',
      TIMESTAMP = '@',
      AUTO_INCREMENT = '+'*/

    switch( desc._type ) {
    case dbase::raw::MEMO:
    case dbase::raw::CHAR:
      return QVariant::String;
    
    case dbase::raw::INTEGER:
      return QVariant::Int;

    case dbase::raw::LOGICAL:
      return QVariant::Bool;

    case dbase::raw::DATE:
      return QVariant::Date;

    case dbase::raw::DATE_TIME:
      return QVariant::DateTime;

    case dbase::raw::TIMESTAMP:
      return QVariant::Time;

    case dbase::raw::NUMBER:
      if( desc._count == 0 ) {
        return QVariant::Int;
      }
      
    case dbase::raw::FLOAT:
    case dbase::raw::DOUBLE:
      return QVariant::Double;

    case dbase::raw::AUTO_INCREMENT:
      return QVariant::LongLong;

    default:
      return QVariant::Invalid;
    }

    return QVariant::Invalid;
  }

  void DBASE::initFields( const std::vector< dbase::raw::FieldDescriptor >& fields ) {
    _fields.clear();

    for(std::vector< dbase::raw::FieldDescriptor >::size_type i=0; i < fields.size(); ++i) {
      const dbase::raw::FieldDescriptor& desc = fields[ i ];
      _fields.push_back( Field( QString::fromStdString( std::string( desc._name ) ), toVariantType( desc ) ) );
    }
  }

  QVariant DBASE::toVariant( const dbase::raw::FieldDescriptor& desc, const QString& d ) const {
    switch( desc._type ) {
     case dbase::raw::LOGICAL:
      {
        /*Y,y 	Yes
        N,n 	No
        F,f 	False
        T,t 	True */
        if( d == "?" ) {
          QVariant b( QVariant::Bool );
          return b;
        }
        QString dUpper = d.toUpper();
        return QVariant( dUpper == "Y" || dUpper == "T" );
      }

      case dbase::raw::DATE:
      {
        if( !d.trimmed().isEmpty() ) {
          bool yok = false;
          bool mok = false;
          bool dok = false;
          int year = d.mid( 0, 4 ).toInt( &yok );
          int month = d.mid( 4, 2 ).toInt( &mok );
          int day = d.mid( 6, 2 ).toInt( &dok );

          if( yok && mok && dok ) {
            return QVariant( QDate( year, month, day ) );
          }
        }
      }
    }

    QString trimmed = d.trimmed();
    QVariant::Type t = toVariantType( desc );
    if( trimmed.isEmpty() ) {
      return QVariant( t );
    }

    if( t == QVariant::String ) {
      return trimmed;
    }
    
    QVariant v( d );    
    if( !v.convert( t ) ) {
      qWarning() << "Datenkonvertierung ist fehlgeschlagen. Rohdaten: '" << d << "' Typ: '" << (char)desc._type << "' -> Zieltyp: '" << QVariant::typeToName( t ) << "'";
      return QVariant();
    }

    return v;
  }
  
  void DBASE::initRecords( const std::vector< dbase::raw::FieldDescriptor >& rawFields, const dbase::raw::Record* records, unsigned int numberOfRecords ) {
    _records.reserve( numberOfRecords );

    QTextCodec* ocodec = QTextCodec::codecForCStrings();
    QTextCodec::setCodecForCStrings( QTextCodec::codecForName( "IBM 850" ) );

    for(unsigned int i=0; i < numberOfRecords; ++i) {
      const dbase::raw::Record& cur = records[ i ];      

      // looking for record state
      if( cur._state != dbase::raw::Record::VALID && cur._state != dbase::raw::Record::DELETED ) {
        qCritical() << "Ungltiger Datensatzzustand: " << (int)cur._state;
        continue;
      }

      // ignore deleted records
      if( cur._state != dbase::raw::Record::VALID ) {
        continue;
      }

      // init record
      Record record( _fields );

      // convert records
      int offset = 0;      
      QString recordData( cur._data );

      QFileInfo info( _fileName );
      QString fileMemo = info.absolutePath() + "/" + info.baseName() + ".dbt";
      QFile memo( fileMemo );

      for(std::vector< dbase::raw::FieldDescriptor >::size_type col = 0; col < rawFields.size(); ++col) {
        const dbase::raw::FieldDescriptor& fieldDesc = rawFields[ col ];

        QString cell = recordData.mid( offset, fieldDesc._length );
        offset += fieldDesc._length;

        // MEMO field handling
        if( fieldDesc._type == dbase::raw::MEMO ) {
          QString tmp = cell.trimmed();
          if( !tmp.isEmpty() ) {
            if( !memo.isOpen() && !memo.open( QIODevice::ReadOnly ) ) {
              qCritical() << "Kann die MEMO Datei '" << fileMemo << "' nicht ffnen.";
            }

            qDebug() << "- Datensatz: " << i << " - Lese MEMO Dateiinhalt fuer Feld: " << QString( fieldDesc._name );

            qint64 memoFileAdress = cell.toUInt() * 512;
            if( !memo.seek( memoFileAdress ) ) {
              qCritical() << "Kann nicht an die MEMO Datei Position springen.";
            } else {
              cell.clear();

              bool readNextBlock = false;
              qint64 rs = 0;
              std::vector< char > t( 512 );
              do {
                rs = memo.read( &t.front(), 512 );

                if( rs == 0 ) {
                  qCritical() << "Kann keine weitere 512 Informationen von Dateilesen.";
                  break;
                }
                
                std::vector< char >::size_type pos = std::numeric_limits< std::vector< char >::size_type >::max();
                for(std::vector< char >::size_type i=0; i < t.size(); ++i) {
                  if( t[i] == 0x1A ) {
                    if( pos >= t.size() ) {
                      pos = i;
                    } else {
                      pos = i-1;
                      break;
                    }
                  }
                }
                    
                readNextBlock = pos >= t.size();

                if( !readNextBlock ) {
                  std::vector< char > sub( pos );
                  memcpy( &sub.front(), &t.front(), sub.size() );
                  sub.push_back( 0x00 );
                  cell += QString( &sub.front() );
                } else {
                  cell += QString( &t.front() );
                }
              } while( readNextBlock );
            }
          }
        }

        record.set( col, toVariant( fieldDesc, cell ) );
      }

      _records.push_back( record );
    }

    QTextCodec::setCodecForCStrings( ocodec );
  }



}