#pragma once

#include <stdint.h>
#include <QtGlobal>
#include <qdatetime.h>
#include <qprogressdialog.h>
#include <qvariant.h>
#include <qfile.h>

namespace dbase {

  namespace raw {

#pragma pack(push, 1)

    enum FieldType : char {
      NO_TYPE = 0,
      CHAR = 'C',
      NUMBER = 'N',
      LOGICAL = 'L',
      DATE = 'D',
      MEMO = 'M',
      FLOAT = 'F',
      // CHAR_NAME_VARIABLE = 'N/A',
      BINARY = 'B',
      GENERAL = 'G',
      PICTURE = 'P',
      CURRENCY = 'Y',
      DATE_TIME = 'T',
      INTEGER = 'I',
      VARI_FIELD = 'V',
      VARIANT = 'X',
      TIMESTAMP = '@',
      DOUBLE = 'O',
      AUTO_INCREMENT = '+'
    };

    struct FieldDescriptor {
      char _name[11];
      FieldType _type;
      uint32_t _dataAdress;
      char _length;
      char _count;
      char _reserved0[2];
      char _workAreaID;
      char _reserved1[2];
      char _flagForSetFields;
      char _reserved[7];
      char _indexFieldFlag;
    };

    struct Header {
      char _version;
      char _lastUpdateYear;
      char _lastUpdateMonth;
      char _lastUpdateDay;
      uint32_t _numberOfRecords;
      uint16_t _headerSize;
      uint16_t _recordSize;
      //char _unused[20];
      char _reserved0[2];
      char _incompleteTransaction;
      char _encriptionFlag; // dBASE IV
      char _multiUserProcessing[12];
      char _mdx;
      char _languageDriverID;
      uint16_t _reserved1; 
    };

    struct Record {
      enum State : char {
        DELETED = 0x2A,
        VALID = 0x20
      };

      State _state;
      char* _data;
    };

    struct MemoHeader {
      uint32_t _numberOfBlocks;
      char _unused[ 508 ];
    };

#pragma pack(pop) 
  }

  struct DBASEInfo {
    char _raw;
    int _versionNumber;
    bool _memoFilePresence;
    bool _sqlTablePresence;
    bool _dbt;

    qint64 _size;
    QDate _lastUpdate;
  };

  class Field {
  public:
    typedef std::vector< Field > Vector;

    Field(const QString& name, QVariant::Type type) 
      : _name( name )
      , _type( type ) 
    {
    }

    const QString& getName() const { return _name; }

    QVariant::Type getType() const { return _type; }

  private:
    QString _name;
    QVariant::Type _type;
  };

  class Record {
  public:
    typedef std::vector< QVariant > QVariantVector;
    Record(const Field::Vector& fields) {
      _vector.resize( fields.size() );
    }

    void set(QVariantVector::size_type col, const QVariant& value) { _vector[ col ] = value; }

    const QVariant& get( QVariantVector::size_type col ) const { return _vector[ col ]; }

  private:
    QVariantVector _vector;
  };


  class DBASE {
  public:
    typedef std::vector< Field > Fields;
    typedef QList< Record > Records;

    static DBASE* readFrom( const QString& file, QProgressDialog* dlg = NULL );

    static bool isEmpty( const QString& file, bool& ok );

    const Fields& getFields() const { return _fields; }

    const Records& getRecords() const { return _records; }

    const DBASEInfo& getInfo() const { return _info; }

    bool isValid() const { return !_fields.empty(); }

    const QString& getFileName() const { return _fileName; }

  protected:
    DBASE( const QString& filename, dbase::raw::Header* header )
      : _fileName( filename )
    {
      initInfo( header );
    }

    DBASE( const QString& filename, dbase::raw::Header* header, std::vector< dbase::raw::FieldDescriptor >& fields, dbase::raw::Record* records )
      : _fileName( filename )
    {
      initInfo( header );

      initFields( fields );

      initRecords( fields, records, header->_numberOfRecords );

      delete[] records;
    }

  private:
    // - for generation -
    static std::auto_ptr< dbase::raw::Header > readHeader( QFile& file );
    static bool readFieldDescriptor( QFile& file, const dbase::raw::Header* header, std::vector< dbase::raw::FieldDescriptor >& res );
    static dbase::raw::Record* readRecords( QFile& file, const dbase::raw::Header* header );
    static bool updateMemoFields( const QString& file, const dbase::raw::Header* header, dbase::raw::Record* );

    //
    void initInfo( const dbase::raw::Header* header );

    void initFields( const std::vector< dbase::raw::FieldDescriptor >& fields );
    QVariant::Type toVariantType( const dbase::raw::FieldDescriptor& desc ) const;

    QVariant toVariant( const dbase::raw::FieldDescriptor& desc, const QString& d ) const;

    void initRecords( const std::vector< dbase::raw::FieldDescriptor >& rawFields, const dbase::raw::Record* records, unsigned int numberOfRecords );

    QString _fileName;
    DBASEInfo _info;
    Fields _fields;
    Records _records;
  };

}